@@ -1,6 +1,7 @@
 const Gdk = imports.gi.Gdk;
 const GLib = imports.gi.GLib;
 const Gio = imports.gi.Gio;
+const GObject = imports.gi.GObject;
 const Gtk = imports.gi.Gtk;
 const Pango = imports.gi.Pango;
 const PangoCairo = imports.gi.PangoCairo;
@@ -20,6 +21,10 @@ const SCROLL_TIMEOUT = 100; // ms
 const TIMESTAMP_INTERVAL = 300; // seconds of inactivity after which to
                                 // insert a timestamp
 
+const INACTIVITY_THRESHOLD = 300; // a threshold in seconds used to control
+                                  // the visibility of status messages
+const STATUS_NOISE_MAXIMUM = 4;
+
 const NUM_INITIAL_LOG_EVENTS = 50; // number of log events to fetch on start
 const NUM_LOG_EVENTS = 10; // number of log events to fetch when requesting more
 
@@ -125,6 +130,86 @@ const TextView = new Lang.Class({
     }
 });
 
+const ButtonTag = new Lang.Class({
+    Name: 'ButtonTag',
+    Extends: Gtk.TextTag,
+    Properties: {
+        'hover': GObject.ParamSpec.boolean('hover',
+                                           'hover',
+                                           'hover',
+                                           GObject.ParamFlags.READWRITE,
+                                           false)
+    },
+    Signals: {
+        'button-press-event': {
+            flags: GObject.SignalFlags.RUN_LAST,
+            param_types: [Gdk.Event.$gtype],
+            return_type: GObject.TYPE_BOOLEAN,
+            accumulator: GObject.AccumulatorType.TRUE_HANDLED
+        },
+        'button-release-event': {
+            flags: GObject.SignalFlags.RUN_LAST,
+            param_types: [Gdk.Event.$gtype],
+            return_type: GObject.TYPE_BOOLEAN,
+            accumulator: GObject.AccumulatorType.TRUE_HANDLED
+        },
+        'clicked': { }
+    },
+
+    _init: function(params) {
+        this.parent(params);
+
+        this._hover = false;
+        this._pressed = false;
+    },
+
+    get hover() {
+        return this._hover;
+    },
+
+    set hover(hover) {
+        if (this._hover == hover)
+            return;
+
+        this._hover = hover;
+        this.notify('hover');
+    },
+
+    on_notify: function(pspec) {
+        if (pspec.name == 'hover' && !this.hover)
+            this._pressed = false;
+    },
+
+    'on_button-press-event': function(event) {
+        let [, button] = event.get_button();
+        this._pressed = button == Gdk.BUTTON_PRIMARY;
+
+        return Gdk.EVENT_STOP;
+    },
+
+    'on_button-release-event': function(event) {
+        let [, button] = event.get_button();
+        if (!(button == Gdk.BUTTON_PRIMARY && this._pressed))
+            return Gdk.EVENT_PROPAGATE;
+
+        this._pressed = false;
+        this.emit('clicked');
+        return Gdk.EVENT_STOP;
+    },
+
+    vfunc_event: function(object, event, iter) {
+        let type = event.get_event_type();
+
+        if (type != Gdk.EventType.BUTTON_PRESS &&
+            type != Gdk.EventType.BUTTON_RELEASE)
+            return Gdk.EVENT_PROPAGATE;
+
+        let isPress = type == Gdk.EventType.BUTTON_PRESS;
+        return this.emit(isPress ? 'button-press-event'
+                                 : 'button-release-event', event);
+    }
+});
+
 const ChatView = new Lang.Class({
     Name: 'ChatView',
 
@@ -137,15 +222,16 @@ const ChatView = new Lang.Class({
         this._onStyleUpdated();
 
         this._room = room;
-        this._state = { lastNick: null, lastTimestamp: 0 };
+        this._state = { lastNick: null, lastTimestamp: 0, lastStatusGroup: 0 };
         this._active = false;
         this._toplevelFocus = false;
         this._joinTime = 0;
         this._maxNickChars = MAX_NICK_CHARS;
-        this._hoveringLink = false;
+        this._hoveredButtonTags = [];
         this._needsIndicator = true;
         this._pending = {};
         this._pendingLogs = [];
+        this._statusCount = { left: 0, joined: 0, total: 0 };
 
         let isRoom = room.type == Tp.HandleType.ROOM;
         let target = new Tpl.Entity({ type: isRoom ? Tpl.EntityType.ROOM
@@ -166,7 +252,7 @@ const ChatView = new Lang.Class({
         this._app = Gio.Application.get_default();
         this._app.pasteManager.addWidget(this._view);
 
-        this._linkCursor = Gdk.Cursor.new(Gdk.CursorType.HAND1);
+        this._hoverCursor = Gdk.Cursor.new(Gdk.CursorType.HAND1);
 
         this._channelSignals = [];
         this._channel = null;
@@ -241,6 +327,9 @@ const ChatView = new Lang.Class({
         context.set_state(Gtk.StateFlags.LINK);
         let linkColor = context.get_color(context.get_state());
         this._activeNickColor = context.get_color(context.get_state());
+
+        context.set_state(Gtk.StateFlags.LINK | Gtk.StateFlags.PRELIGHT);
+        this._hoveredLinkColor = context.get_color(context.get_state());
         context.restore();
 
         let desaturatedNickColor = (this._activeNickColor.red +
@@ -253,6 +342,12 @@ const ChatView = new Lang.Class({
         if (this._activeNickColor.equal(this._inactiveNickColor))
             this._inactiveNickColor.alpha = 0.5;
 
+        context.save();
+        context.add_class('view');
+        context.set_state(Gtk.StateFlags.NORMAL);
+        this._statusHeaderHoverColor = context.get_color(context.get_state());
+        context.restore();
+
         let buffer = this._view.get_buffer();
         let tagTable = buffer.get_tag_table();
         let tags = [
@@ -290,6 +385,9 @@ const ChatView = new Lang.Class({
                                     wrap_mode: Gtk.WrapMode.WORD_CHAR,
                                     right_margin: MARGIN });
 
+        this._view.add_events(Gdk.EventMask.LEAVE_NOTIFY_MASK);
+        this._view.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK);
+
         this.widget.add(this._view);
         this.widget.show_all();
 
@@ -307,12 +405,13 @@ const ChatView = new Lang.Class({
                                         Lang.bind(this, this._onValueChanged));
         this.widget.vadjustment.connect('changed',
                                  Lang.bind(this, this._updateScroll));
-        this._view.connect('button-release-event',
-                           Lang.bind(this, this._handleLinkClicks));
-        this._view.connect('button-press-event',
-                           Lang.bind(this, this._handleLinkClicks));
+        this._view.connect('key-press-event', Lang.bind(this, this._onKeyPress));
         this._view.connect('motion-notify-event',
-                           Lang.bind(this, this._handleLinkHovers));
+                           Lang.bind(this, this._handleButtonTagsHover));
+        this._view.connect('enter-notify-event',
+                           Lang.bind(this, this._handleButtonTagsHover));
+        this._view.connect('leave-notify-event',
+                           Lang.bind(this, this._handleButtonTagsHover));
     },
 
     _onDestroy: function() {
@@ -426,7 +525,8 @@ const ChatView = new Lang.Class({
         let adj = this.widget.vadjustment;
         if (adj.value == this._scrollBottom) {
             if (this._nPending == 0) {
-                adj.value = adj.upper - adj.page_size;
+                this._view.emit('move-cursor',
+                                Gtk.MovementStep.BUFFER_ENDS, 1, false);
             } else {
                 let id = Object.keys(this._pending).sort(function(a, b) {
                     return a - b;
@@ -446,6 +546,21 @@ const ChatView = new Lang.Class({
         if (hasDeltas && dy >= 0)
             return Gdk.EVENT_PROPAGATE;
 
+        return this._fetchBacklog();
+    },
+
+    _onKeyPress: function(w, event) {
+        let [, keyval] = event.get_keyval();
+        if (keyval != Gdk.KEY_Up &&
+            keyval != Gdk.KEY_KP_Up &&
+            keyval != Gdk.KEY_Page_Up &&
+            keyval != Gdk.KEY_KP_Page_Up)
+            return Gdk.EVENT_PROPAGATE;
+
+        return this._fetchBacklog();
+    },
+
+    _fetchBacklog: function() {
         if (this.widget.vadjustment.value != 0 ||
             this._logWalker.is_end())
             return Gdk.EVENT_PROPAGATE;
@@ -505,61 +620,36 @@ const ChatView = new Lang.Class({
         menu.popup(null, null, null, button, time);
     },
 
-    _handleLinkClicks: function(view, event) {
-        let isPress = event.get_event_type() == Gdk.EventType.BUTTON_PRESS;
-
-        let [, button] = event.get_button();
-        if (button != Gdk.BUTTON_PRIMARY &&
-            (button != Gdk.BUTTON_SECONDARY || !isPress))
-            return Gdk.EVENT_PROPAGATE;
-
-        if (isPress)
-            this._clickedUrl = null;
-
-        let [, eventX, eventY] = event.get_coords();
-        let [x, y] = view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET,
-                                                  eventX, eventY);
-
-        let iter = view.get_iter_at_location(x, y);
-        let tags = iter.get_tags();
-        for (let i = 0; i < tags.length; i++) {
-            let url = tags[i]._url;
-            if (url) {
-                if (url.indexOf(':') == -1)
-                    url = 'http://' + url;
-
-                if (isPress) {
-                    if (button == Gdk.BUTTON_PRIMARY)
-                        this._clickedUrl = url;
-                    else
-                        this._showUrlContextMenu(url, button, event.get_time());
-                    return Gdk.EVENT_STOP;
-                } else if (this._clickedUrl == url) {
-                    Utils.openURL(url, event.get_time());
-                    return Gdk.EVENT_STOP;
-                }
-                break;
-            }
-        }
-        return Gdk.EVENT_PROPAGATE;
-    },
-
-    _handleLinkHovers: function(view, event) {
+    _handleButtonTagsHover: function(view, event) {
         let [, eventX, eventY] = event.get_coords();
         let [x, y] = view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET,
                                                   eventX, eventY);
         let iter = view.get_iter_at_location(x, y);
-        let tags = iter.get_tags();
-        let hovering = false;
-        for (let i = 0; i < tags.length && !hovering; i++)
-            if (tags[i]._url)
-                hovering = true;
 
-        if (this._hoveringLink != hovering) {
-            this._hoveringLink = hovering;
-            let cursor = this._hoveringLink ? this._linkCursor : null;
+        let hoveredButtonTags = iter.get_tags().filter(
+            function(t) {
+                return t instanceof ButtonTag;
+            });
+
+        hoveredButtonTags.forEach(
+            function(t) {
+                t.hover = true;
+            });
+        this._hoveredButtonTags.forEach(
+            function(t) {
+                t.hover = hoveredButtonTags.indexOf(t) >= 0;
+            });
+
+        let isHovering = hoveredButtonTags.length > 0;
+        let wasHovering = this._hoveredButtonTags.length > 0;
+
+        if (isHovering != wasHovering) {
+            let cursor = isHovering ? this._hoverCursor : null;
             this._view.get_window(Gtk.TextWindowType.TEXT).set_cursor(cursor);
         }
+
+        this._hoveredButtonTags = hoveredButtonTags;
+
         return Gdk.EVENT_PROPAGATE;
     },
 
@@ -651,9 +741,9 @@ const ChatView = new Lang.Class({
 
         let channelSignals = [
             { name: 'message-received',
-              handler: Lang.bind(this, this._insertTpMessage) },
+              handler: Lang.bind(this, this._onMessageReceived) },
             { name: 'message-sent',
-              handler: Lang.bind(this, this._insertTpMessage) },
+              handler: Lang.bind(this, this._onMessageSent) },
             { name: 'pending-message-removed',
               handler: Lang.bind(this, this._pendingMessageRemoved) }
         ];
@@ -681,8 +771,8 @@ const ChatView = new Lang.Class({
     },
 
     _onMemberRenamed: function(room, oldMember, newMember) {
-        this._insertStatus(_("%s is now known as %s").format(oldMember.alias,
-                                                             newMember.alias));
+        let text = _("%s is now known as %s").format(oldMember.alias, newMember.alias);
+        this._insertStatus(text, oldMember.alias, 'renamed');
         this._setNickStatus(oldMember.alias, Tp.ConnectionPresenceType.OFFLINE);
         this._setNickStatus(newMember.alias, Tp.ConnectionPresenceType.AVAILABLE);
     },
@@ -691,7 +781,7 @@ const ChatView = new Lang.Class({
         let text = _("%s has disconnected").format(member.alias);
         if (message)
             text += ' (%s)'.format(message);
-        this._insertStatus(text);
+        this._insertStatus(text, member.alias, 'left');
         this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
     },
 
@@ -700,7 +790,7 @@ const ChatView = new Lang.Class({
             actor ? _("%s has been kicked by %s").format(member.alias,
                                                          actor.alias)
                   : _("%s has been kicked").format(member.alias);
-        this._insertStatus(message);
+        this._insertStatus(message, member.alias, 'left');
         this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
     },
 
@@ -709,31 +799,145 @@ const ChatView = new Lang.Class({
             actor ? _("%s has been banned by %s").format(member.alias,
                                                          actor.alias)
                   : _("%s has been banned").format(member.alias)
-        this._insertStatus(message);
+        this._insertStatus(message, member.alias, 'left');
         this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
     },
 
     _onMemberJoined: function(room, member) {
-        this._insertStatus(_("%s joined").format(member.alias));
+        let text = _("%s joined").format(member.alias);
+        this._insertStatus(text, member.alias, 'joined');
         this._setNickStatus(member.alias, Tp.ConnectionPresenceType.AVAILABLE);
     },
 
     _onMemberLeft: function(room, member, message) {
         let text = _("%s left").format(member.alias);
+
         if (message)
             text += ' (%s)'.format(message);
-        this._insertStatus(text);
+
+        this._insertStatus(text, member.alias, 'left');
         this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
     },
 
-    _insertStatus: function(text) {
+    _onMessageReceived: function(room, tpMessage) {
+        this._insertTpMessage(room, tpMessage);
+        this._resetStatusCompressed();
+        let nick = tpMessage.sender.alias;
+        let nickTag = this._lookupTag('nick' + nick);
+        if (!nickTag)
+           return;
+        nickTag._lastActivity = GLib.get_monotonic_time();
+    },
+
+    _onMessageSent: function(room, tpMessage) {
+        this._insertTpMessage(room, tpMessage);
+        this._resetStatusCompressed();
+    },
+
+    _resetStatusCompressed: function() {
+        let markStart = this._view.buffer.get_mark('idle-status-start');
+        if (!markStart)
+            return;
+
+        this._view.buffer.delete_mark(markStart);
+        this._statusCount =  { left: 0, joined: 0, total: 0 };
+        this._state.lastStatusGroup++;
+    },
+
+    _shouldShowStatus: function(nick) {
+        let nickTag = this._lookupTag('nick' + nick);
+
+        if (!nickTag)
+            return false;
+
+        let time = GLib.get_monotonic_time();
+        return (time - nickTag._lastActivity) / (1000 * 1000) < INACTIVITY_THRESHOLD;
+    },
+
+    _updateStatusHeader: function() {
+        let buffer = this._view.buffer;
+        let headerMark = buffer.get_mark('idle-status-start');
+
+        let headerTagName = 'status-compressed' + this._state.lastStatusGroup;
+        let groupTagName = 'status' + this._state.lastStatusGroup;
+
+        let headerTag, groupTag;
+        if (!headerMark) {
+            // we are starting a new group
+            headerTag = new ButtonTag({ name: headerTagName, invisible: true });
+            groupTag = new Gtk.TextTag({ name: groupTagName });
+            buffer.tag_table.add(headerTag);
+            buffer.tag_table.add(groupTag);
+
+            headerTag.connect('clicked',
+                function() {
+                    groupTag.invisible = !groupTag.invisible;
+                });
+
+            headerTag.connect('notify::hover', Lang.bind(this,
+                function() {
+                    headerTag.foreground_rgba = headerTag.hover ? this._statusHeaderHoverColor : null;
+                }));
+
+            this._ensureNewLine();
+            headerMark = buffer.create_mark('idle-status-start', buffer.get_end_iter(), true);
+        } else {
+            headerTag = this._lookupTag(headerTagName);
+            groupTag = this._lookupTag(groupTagName);
+
+            let start = buffer.get_iter_at_mark(headerMark);
+            let end = start.copy();
+            end.forward_to_line_end(headerTag);
+            buffer.delete(start, end);
+        }
+
+        // we passed the threshold, show the header and collapse the group
+        if (this._statusCount.total > STATUS_NOISE_MAXIMUM && headerTag.invisible) {
+            headerTag.invisible = false;
+            groupTag.invisible = true;
+        }
+
+        let stats = [];
+        if (this._statusCount.joined > 0)
+            stats.push(ngettext("%d user joined",
+                                "%d users joined", this._statusCount.joined).format(this._statusCount.joined));
+        if (this._statusCount.left > 0)
+            stats.push(ngettext("%d user left",
+                                "%d users left", this._statusCount.left).format(this._statusCount.left));
+        this._insertWithTags(buffer.get_iter_at_mark(headerMark),
+                             stats.join(", ") + ' (\u2026)',
+                             [this._lookupTag('status'), headerTag]);
+    },
+
+    _insertStatus: function(text, member, type) {
         let time = GLib.DateTime.new_now_utc().to_unix();
         if (time - this._joinTime < IGNORE_STATUS_TIME)
             return;
+
+        let grouped = time - this._state.lastTimestamp > INACTIVITY_THRESHOLD;
+        if (!grouped && !this._shouldShowStatus(member))
+            return;
+
         this._state.lastNick = null;
+
+        let tags = [this._lookupTag('status')];
+        let groupTag = null;
+        if (grouped) {
+            if (this._statusCount.hasOwnProperty(type)) {
+                this._statusCount[type]++;
+                this._statusCount.total++;
+            }
+            this._updateStatusHeader();
+
+            groupTag = this._lookupTag('status' + this._state.lastStatusGroup);
+            tags.push(groupTag);
+        } else {
+            this._resetStatusCompressed();
+        }
+
         this._ensureNewLine();
         let iter = this._view.buffer.get_end_iter();
-        this._insertWithTagName(iter, text, 'status');
+        this._insertWithTags(iter, text, tags);
     },
 
     _formatTimestamp: function(timestamp) {
@@ -942,8 +1146,7 @@ const ChatView = new Lang.Class({
             let url = urls[i];
             this._insertWithTags(iter, text.substr(pos, url.pos - pos), tags);
 
-            let tag = new Gtk.TextTag();
-            tag._url = url.url;
+            let tag = this._createUrlTag(url.url);
             this._view.get_buffer().tag_table.add(tag);
 
             this._insertWithTags(iter, url.url,
@@ -954,11 +1157,43 @@ const ChatView = new Lang.Class({
         this._insertWithTags(iter, text.substr(pos), tags);
     },
 
+    _createUrlTag: function(url) {
+        if (url.indexOf(':') == -1)
+            url = 'http://' + url;
+
+        let tag = new ButtonTag();
+        tag.connect('notify::hover', Lang.bind(this,
+            function() {
+                tag.foreground_rgba = tag.hover ? this._hoveredLinkColor : null;
+            }));
+        tag.connect('clicked',
+            function() {
+                Utils.openURL(url, Gtk.get_current_event_time());
+            });
+        tag.connect('button-press-event', Lang.bind(this,
+            function(tag, event) {
+                let [, button] = event.get_button();
+                if (button != Gdk.BUTTON_SECONDARY)
+                    return Gdk.EVENT_PROPAGATE;
+
+                this._showUrlContextMenu(url, button, event.get_time());
+                return Gdk.EVENT_STOP;
+            }));
+        return tag;
+    },
+
     _ensureNewLine: function() {
         let buffer = this._view.get_buffer();
         let iter = buffer.get_end_iter();
+        let tags = [];
+        let groupTag = this._lookupTag('status' + this._state.lastStatusGroup);
+        if (groupTag && iter.ends_tag(groupTag))
+            tags.push(groupTag);
+        let headerTag = this._lookupTag('status-compressed' + this._state.lastStatusGroup);
+        if (headerTag && iter.ends_tag(headerTag))
+            tags.push(headerTag);
         if (iter.get_line_offset() != 0)
-            buffer.insert(iter, '\n', -1);
+            this._insertWithTags(iter, '\n', tags);
     },
 
     _getLineIters: function(iter) {
