@@ -197,7 +197,11 @@ const MainWindow = new Lang.Class({
         // Make sure user-list button is at least as wide as icon buttons
         this._joinMenuButton.connect('size-allocate', Lang.bind(this,
             function(w, rect) {
-                this._showUserListButton.width_request = rect.width;
+                let width = rect.width;
+                Mainloop.idle_add(Lang.bind(this, function() {
+                    this._showUserListButton.width_request = width;
+                    return GLib.SOURCE_REMOVE;
+                }));
             }));
 
         let scroll = builder.get_object('room_list_scrollview');
@@ -212,20 +216,12 @@ const MainWindow = new Lang.Class({
         let dialog = new JoinDialog.JoinDialog();
         dialog.widget.transient_for = this.window;
         dialog.widget.show();
-        dialog.widget.connect('response',
-            function(widget) {
-                widget.destroy();
-            });
     },
 
     showMessageUserDialog: function() {
         let dialog = new MessageDialog.MessageDialog();
         dialog.widget.transient_for = this.window;
         dialog.widget.show();
-        dialog.widget.connect('response',
-            function(widget) {
-                widget.destroy();
-            });
     },
 
     _updateUserListLabel: function() {
