@@ -120,7 +120,8 @@ const RoomRow = new Lang.Class({
     },
 
     _createWidget: function(gicon) {
-        this.widget = new Gtk.ListBoxRow({ margin_bottom: 4 });
+        this.widget = new Gtk.ListBoxRow({ margin_bottom: 4,
+                                           focus_on_click: false });
 
         this._eventBox = new Gtk.EventBox();
         this.widget.add(this._eventBox);
@@ -158,6 +159,7 @@ const RoomRow = new Lang.Class({
 const RoomListHeader = new Lang.Class({
     Name: 'RoomListHeader',
     Extends: Gtk.MenuButton,
+    CssName: 'row',
     Template: 'resource:///org/gnome/Polari/room-list-header.ui',
     InternalChildren: ['label',
                        'iconStack',
@@ -175,8 +177,6 @@ const RoomListHeader = new Lang.Class({
 
         this.parent(params);
 
-        this.get_style_context().remove_class('button');
-
         this._errorPopover.relative_to = this._iconStack;
         this._popoverButton.connect('clicked', Lang.bind(this,
             function() {
@@ -296,6 +296,7 @@ const RoomList = new Lang.Class({
 
     _init: function() {
         this.widget = new Gtk.ListBox({ hexpand: false });
+        this.widget.get_style_context().add_class('sidebar');
 
         this.widget.set_selection_mode(Gtk.SelectionMode.BROWSE);
         this.widget.set_header_func(Lang.bind(this, this._updateHeader));
